//==========================================================================
// 
// Copyright 2001 The Regents of the University of California 
// All Rights Reserved 
// 
// Permission to use, copy, modify and distribute any part of this skitterTrace
// for educational, research and non-profit purposes, without fee, and without 
// a written agreement is hereby granted, provided that the above copyright 
// notice, this paragraph and the following paragraphs appear in all copies. 
// 
// Those desiring to incorporate this into commercial products or use for 
// commercial purposes should contact the 
//
//	Technology Transfer Office, University of California, San Diego, 
//	9500 Gilman Drive, La Jolla, CA 92093-0910
//	Ph: (619) 534-5815, FAX: (619) 534-7345. 
// 
// IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
// DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING
// LOST PROFITS, ARISING OUT OF THE USE OF THIS SK_ANALYSIS_DUMP, EVEN IF THE 
// UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// THE SK_ANALYSIS_DUMP, PROVIDED HEREIN IS ON AN "AS IS" BASIS, AND THE 
// UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT,
// UPDATES, ENHANCEMENTS, OR MODIFICATIONS. THE UNIVERSITY OF CALIFORNIA MAKES
// NO REPRESENTATIONS AND EXTENDS NO WARRANTIES OF ANY KIND, EITHER IMPLIED OR
// EXPRESS, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
// MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE, OR THAT THE USE OF THE
// SK_ANALYSIS_DUMP WILL NOT INFRINGE ANY PATENT, TRADEMARK OR OTHER RIGHTS. 
//
// sk_analysis_dump is developed by Bradley L. Huffaker at the University of
// California, San Diego under the Cooperative Association for
// Internet Data Analysis (CAIDA) Program. Support for this effort is
// provided by DARPA grant N66001-98-2-8922 and by CAIDA members.
//==========================================================================
//
// This program prints out a perl friendly version of ArtsIpPathData.
// In order to do this is does some simple preanalysis.
//      N - no reply
//              There was no response from the Final destination
//              and as such any RTT value must be ignored
//      I - incomplete
//              A valid response and RTT was collected for the final
//              destination, but some TTL IPs are missing or more than
//              one IP responsed to a same TTL.
//      C - complete
//              A valid response and RTT was collected for the final
//              destination and all TTLS have a single IP address.

#include <fstream.h>
#include <stdio.h>
#include <set.h>

#include "Arts.hh"
#include "caida_t.h"
#include "config.h"

#ifdef USE_ZFSTREAM_
#include "zfstream.h"
#endif

#ifdef USE_GEODE_
#include "Geode.hh"
#endif

    struct geo_info_struct {
	int print_geo_info;
	geo_info_struct() {
	    print_geo_info = 0;
	}

#ifdef USE_GEODE_

	Geode geode;
	set<ipv4addr_t> seen_ip;
	
	bool Open(const char *geodeHostname, int port = 9292) {
	    return geode.Open(geodeHostname, port);
	}
#endif
    };

#define MAX_NUM_HOPS	300
#define MAX_NUM_IPS	100

#define MAX_NUM_HOPS_PRINT_ERROR  255
#define MAX_NUM_IPS_PRINT_ERROR   3

#define MAX_NUM_SKIPPED_HOPS	5

#define	FORMAT_DEFAULT	0
#define	FORMAT_SIMPLE	1
#define	FORMAT_DATABASE	2


//-------------------------------------------------------------------------
//                      void Usage(const char *argv0)
//.........................................................................
//
//-------------------------------------------------------------------------
void Usage(const char *argv0) {
    cerr << "Usage: " << argv0;
#ifdef USE_GEODE_
    cerr << " [-shgd] [-G server]";
#endif
#ifndef USE_GEODE_
    cerr << " [-shd]";
#endif
    cerr << " [-S numlines] [-D numlines] file(s)" << endl;
    return;
}

//-------------------------------------------------------------------------
//                      void PrintHelp()
//.........................................................................
//
//-------------------------------------------------------------------------
void PrintHelp() {
    cerr << "  default behavior - most detailed output\n";
    cerr << "      all IP addresses seen for a given hop are printed\n";
    cerr << "      separated by commas\n";
    cerr << "      IP seen at hop counts greater than the one for the\n";
    cerr << "      destination are printed\n";
    cerr << "      q is printed for hops when no IP address was found\n";
    cerr << endl;

    cerr << "  s - short format\n";
    cerr << "      included for backwards compatablity with older versions\n";
    cerr << "      only the first IP address seen for a given hop is printed\n";
    cerr << "      IP seen at hop counts greater than the one for the\n";
    cerr << "      destination are not printed\n";
    cerr << "      note: use of this options is discouraged\n";
    cerr << endl;
    
    cerr << "  b - database format\n";
    cerr << "      replaces all the whitespace with tabs\n";
    cerr << "      places the destination where it belongs in the path\n";
    cerr << "      if it should be there\n";
    cerr << endl;

    cerr << "  D numline - debug mode that only reads the first numline\n";
    cerr << "      arts objects\n";
    cerr << "  S numline - skips the first numline lines in the file\n";
    cerr << endl;

#ifdef USE_GEODE_
    cerr << "  g - print out geographical information\n";
    cerr << "      assuming that environmental variable GEODESERVER is set\n";
    cerr << "      to the GEODE server\n";
    cerr << "  G servername - the same as g except it uses the servername\n";
    cerr << "      given on the command line.\n";
    cerr << endl;
#endif

    cerr << "  h - prints this message\n";
    cerr << endl;
    return;
}


//-------------------------------------------------------------------------
//			void ProcessArts(Arts * arts);
//.........................................................................
//
//            Process the Arts object and prints out it's content
//-------------------------------------------------------------------------
void ProcessStream(bool debug_mode, int debug_mode_numlines, int skip_numlines,
    char * argv0, char * stream_name, 
    int format, geo_info_struct * geo_info, istream * stream);
void ProcessArts(char * stream_name, int format, 
  geo_info_struct * geo_info, Arts * arts);

void PrintGeoInfo(FILE * output, int format, geo_info_struct * geo_info, 
    ipv4addr_t ip);

//-------------------------------------------------------------------------
//                     int main(int argc, char *argv[])                    
//.........................................................................
//  
//-------------------------------------------------------------------------
int main(int argc, char *argv[]) {

    int format = FORMAT_DEFAULT;
    geo_info_struct geo_info;
  
    int opt;
    
    char *geodeHostname = getenv("GEODESERVER");

    bool debug_mode = false;
    int debug_mode_numlines;
    int skip_numlines = -1;
    
    while ((opt = getopt(argc, argv, "hdsgS:D:G:")) >= 0) {
	if (opt == 's') {
	    format = FORMAT_SIMPLE;
	} else if (opt == 'd') {
	    format = FORMAT_DATABASE;

	} else if (opt == 'S') {
	    skip_numlines = atoi(optarg);
	} else if (opt == 'D') {
	    debug_mode = true;
	    debug_mode_numlines = atoi(optarg);

#ifdef USE_GEODE_
	} else if (opt == 'g') {
	    geo_info.print_geo_info = 1;
	} else if (opt == 'G') {
	    geo_info.print_geo_info = 1;
            geodeHostname =  new char[strlen(optarg)+1];
            strcpy(geodeHostname, optarg);
#endif
	} else if (opt == 'h') {
	    Usage(argv[0]);
	    PrintHelp();	
	    exit(0);
	} else {
	    Usage(argv[0]);
	    exit(-1);
	}
    }

#ifdef USE_GEODE_
    if (geo_info.print_geo_info == 1 && !geo_info.Open(geodeHostname)) {
	cerr << "Unable to connect to Geode Server\n";
	exit(1);
    }
#endif


    switch(format) {
	case FORMAT_SIMPLE:
	    printf ("#key src dst time_stamp rtt hop_count hop1 hop2 . . .\n");
	    break;
	case FORMAT_DATABASE:
	    printf ("#key\tsrc\tdst\ttime_stamp\trtt\thop1\thop2 . . .\n");
	    printf ("#   tab delaminated\n");
	    break;
	default :
	    printf ("#key src dst time_stamp rtt hop1 hop2 . . .\n");
    }
    printf ("#  key\n");
    printf ("#-------\n");
    printf ("#   no reply\n");
    printf ("#    N - no reply\n");
    printf ("#   replied\n");
    printf ("#    I - incomplete\n");
    printf ("#    C - complete\n");
    switch(format) {
	case FORMAT_SIMPLE:
	    printf ("#   hop* - can contain one or more comma separated IP"); 
	    break;
	default :
	    printf ("#   hop* - contains a single IP");
    }
    printf (" or a `q' for missing hops\n");
#ifdef USE_GEODE_
    if (geo_info.print_geo_info == 1) {
	printf ("#   geographical\n");
	printf ("#    G ip city:state:country:lat:long\n");
	printf ("#	city,state,country will be unknown if not provided\n");
	printf ("#	lat,long will both be 0 if they are not provided\n");
    }
#endif
	

    if (optind < argc) {
	for (int fileNum = optind; fileNum < argc; fileNum++) {
#ifdef USE_ZFSTREAM_
	    istream * artsStream = new gzifstream(argv[fileNum]);
#else
	    ifstream * artsStream = new ifstream(argv[fileNum]);
#endif
	    if (! artsStream || ! *artsStream) 
	    {
		cerr << "unable to open " << argv[fileNum] << endl;
		cout << "# unable to open " << argv[fileNum] << endl;
	    } else {
		ProcessStream(debug_mode, debug_mode_numlines, skip_numlines,
		    argv[0], argv[fileNum], format, 
		    &geo_info, artsStream);
		//  delete the input stream.
		delete(artsStream);
	    }
	}
    } else {
	ProcessStream(debug_mode, debug_mode_numlines, skip_numlines,
	    argv[0], "STDIN", format, &geo_info, &cin);
    }
}

void ProcessStream(bool debug_mode, int debug_mode_numlines, int skip_numlines,
    char * argv0, char * stream_name, 
    int format, geo_info_struct * geo_info, istream * stream) {
    Arts            *arts;
    int num_found = 0;
    while (*stream && (!debug_mode || num_found < debug_mode_numlines ) ) {
	arts = new Arts();

	//  read an Arts object.

	arts->read(*stream);
	if (! *stream) {
	    //  error or EOF while reading
	    delete(arts);
	    break;
	}
	num_found++;
	if (num_found > skip_numlines) 
	    ProcessArts(stream_name, format, geo_info, arts);
	//  delete the object.
	delete(arts);
    }
    printf("#END %s DATA %s\n", argv0,  stream_name);
}

void ProcessArts(char * stream_name, int format, 
  geo_info_struct * geo_info, Arts * arts) {
    //  Skip objects that aren't IP path objects.
    if (arts->Header().Identifier() != artsC_OBJECT_IP_PATH) {
	return;
    }

    // Get a pointer to the path data stored in the Arts object.
    ArtsIpPathData * ipPathData = arts->IpPathData();

    //  Do whatever you want with ipPathData using members of the
    //  ArtsIpPathData class.  For example, print the RTT if the
    //  destination was reached.

    // The following checks will provide you with only complete
    // paths where the destination and all the intermediate hops
    // were also reached.
    // IsComplete is not well named. It only means that the 
    // destination was reached.

    vector<ArtsAttribute>::const_iterator creationAttribute
	= arts->FindCreationAttribute();
    if (creationAttribute != arts->Attributes().end())
    {
	vector<ArtsIpPathEntry> & pathHops = ipPathData->Path();
	vector<ArtsIpPathEntry>::const_iterator hopIter;

	int last_hop_seen = 0;

	int num_ips[MAX_NUM_HOPS];
	ipv4addr_t ip_per_hop[MAX_NUM_HOPS][MAX_NUM_IPS];
	for (int i=0; i < MAX_NUM_HOPS; i++) {
	   num_ips[i] = 0;
	}

	if (geo_info->print_geo_info ) {
	    PrintGeoInfo(stdout, format, geo_info, ipPathData->Src());
	    PrintGeoInfo(stdout, format, geo_info, ipPathData->Dst());
	}

	// This loops has two purposes:
	//	it sorts the hops which is not guaranteed by skitter
	//	it also organized IP by hops so they can be printed
	//	together
	for (hopIter = pathHops.begin(); hopIter != pathHops.end(); hopIter++) {
	    in_addr addr;
	    addr.s_addr = ipPathData->Dst();
	    int hop = (*hopIter).HopNum();
	    if (hop >= MAX_NUM_HOPS) {
		fprintf (stderr, 
		    "Found in `%s' for destination `%s' more than %d hops\n", 
		    stream_name, inet_ntoa(addr), MAX_NUM_HOPS);
		return;
	    }
	    if (num_ips[hop] >= MAX_NUM_IPS) {
		fprintf (stderr, 
		    "Found in `%s' for destination `%s' more than %d ips\n", 
		    stream_name, inet_ntoa(addr), MAX_NUM_IPS);
		return;
	    }
	    ipv4addr_t ip = (*hopIter).IpAddr();
	    ip_per_hop[hop][num_ips[hop]] = ip;
	    if (geo_info->print_geo_info ) {
		PrintGeoInfo(stdout, format, geo_info, ip);
	    }
		
	    num_ips[hop]++;
	    if (last_hop_seen < hop) 
		last_hop_seen = hop;
	}

	const long int sample_time = (*creationAttribute).Creation();
	int dst_hop = 0;
	double total_rtt = 0;
	if (ipPathData->IsComplete()) {
	    dst_hop = ipPathData->HopDistance();
	    total_rtt = (ipPathData->Rtt().tv_sec * 1000.0) +
		(ipPathData->Rtt().tv_usec / 1000.0);
	}

	
	int hop = 1;
	while (num_ips[hop] > 0 && hop < dst_hop) {
	    hop++;
	}
    
	char key;
	if (!ipPathData->IsComplete()) {
	    key = 'N';
	} else {
	    if (hop < dst_hop) 
		key = 'I';
	    else
		key = 'C';
	}

	in_addr addr;

	char delaminator;
	switch(format) {
	    case FORMAT_DATABASE:
		delaminator = '\t';
		break;
	    default:
		delaminator = ' ';
	}
	addr.s_addr = ipPathData->Src();
	printf ("%c%c%s", key, delaminator, inet_ntoa(addr));
	addr.s_addr = ipPathData->Dst();
	printf ("%c%s", delaminator, inet_ntoa(addr));

	printf ("%c%lu%c%.3f", delaminator,sample_time, delaminator, total_rtt);
        if (FORMAT_SIMPLE != format)
	    printf ("%c%d", delaminator, ipPathData->HopDistance());

	// print extra q for missing tails
	// also if not printing ip_count, don't print extra IP at the end
	if (ipPathData->IsComplete()) {
	    if (FORMAT_DATABASE == format 
		    && last_hop_seen+MAX_NUM_SKIPPED_HOPS >= dst_hop) {
		ip_per_hop[dst_hop][num_ips[dst_hop]++] = ipPathData->Dst();
		if (last_hop_seen < dst_hop) 
		    last_hop_seen = dst_hop;
	    } else if (FORMAT_SIMPLE == format) {
		last_hop_seen = dst_hop-1;
	    }
	}
	for (int hop=1; hop <= last_hop_seen; hop++) {
	    // If the simple_format is set is false and more than one IP were
	    // found for a given hop, that hop must be left undecided
	    if (num_ips[hop] == 0 || 
		((FORMAT_SIMPLE == format) && num_ips[hop] > 1)) {
		printf ("%cq", delaminator);
	    } else {
		char lead_char =  delaminator;
		int ip_index = 0;
		while (ip_index < num_ips[hop]) { 
		    in_addr addr;
		    addr.s_addr = ip_per_hop[hop][ip_index] ;
		    printf ("%c%s", lead_char,inet_ntoa( addr),hop );
		    lead_char = ',';
		    ip_index++;
		}
	    }
	}
	printf ("\n");
    }
}

void PrintGeoInfo(FILE * output, int format, geo_info_struct * geo_info, 
    ipv4addr_t ip) {
#ifdef USE_GEODE_

    if (geo_info->seen_ip.find(ip) != geo_info->seen_ip.end())
	return;
    geo_info->seen_ip.insert(ip);

    char delaminator;
    switch(format) {
	case FORMAT_DATABASE:
	    delaminator = '\t';
	    break;
	default:
	    delaminator = ' ';
    }

    uint num_times_tried = 0;

    const GeodeError * geodeError = (GeodeError*)0;
    const char * unknown = "unknown";
    const char * city = unknown;
    const char * state = unknown;
    const char * country = unknown;
    float latitude = 0;
    float longitude = 0;

    GeodeReply * reply = geo_info->geode.GetRecord(ip);
    if (reply) {
	uint32_t  index = reply->Record().Index();
	if ( (reply->Record().k_latitudeMask &index) != 0
	  && (reply->Record().k_longitudeMask & index ) != 0) {
	    latitude = reply->Record().Latitude();
	    longitude = reply->Record().Longitude();
	    if ((reply->Record().k_cityMask &index) != 0)
		city = reply->Record().City();
	    if ((reply->Record().k_stateMask &index) != 0)
		state = reply->Record().State();
	    if ((reply->Record().k_countryCodeMask &index) != 0)
		country = reply->Record().CountryCodeIso2();
	    
	}
    } else {

	geodeError = geo_info->geode.LastGetError();
	if (geodeError == (GeodeError *)0) {
	    cout << "No reply and no error message\n";
	    exit(0);
	}
    }
    in_addr addr;
    addr.s_addr = ip;
    fprintf( output, "G%c%s%c%s:%s:%s:%.2f:%.2f\n", 
	delaminator, inet_ntoa(addr), delaminator, 
	city,state,country,latitude,longitude);
    if (reply)
	    delete reply;
#endif
}
